# file generated by setuptools_scm
# don't change, don't track in version control
__version__ = version = '0.0.21.dev7+g99ff450.d20230409'
__version_tuple__ = version_tuple = (0, 0, 21, 'dev7', 'g99ff450.d20230409')
